<?php

declare(strict_types=1);

namespace Drupal\violinist_projects;

/**
 * An analyzer of logs.
 */
final class LogAnalyzer implements LogAnalyzerInterface {

  /**
   * {@inheritdoc}
   */
  public function getUpdateStatusFromOutput(string $console_output) : UpdateResultValue {
    // First replace all weird newlines with normal ones.
    $console_output = str_replace(["\r\n", "\r"], "\n", $console_output);
    // Then, any double new lines should go away.
    $console_output = preg_replace("/\n\n+/", "\n", $console_output);
    // Then, we can split the lines.
    $lines = explode("\n", $console_output);
    // Now let's inspect all of them, and if we find something we know about,
    // return that as the reason.
    foreach ($lines as $line) {
      if (str_contains($line, 'is blocked by your allow-plugins config')) {
        return UpdateResultValue::NewPluginNotAllowed;
      }
      if (str_contains($line, 'Cannot apply patch') || str_contains($line, 'In Patches.php line ')) {
        return UpdateResultValue::CannotApplyPatch;
      }
    }
    return UpdateResultValue::Unknown;
  }

  /**
   * {@inheritdoc}
   */
  public function getUpdateStatusHumanReadable(string $console_output): string {
    $status = $this->getUpdateStatusFromOutput($console_output);
    switch ($status) {
      case UpdateResultValue::NewPluginNotAllowed:
        return 'The update probably contains installing a plugin which was not there before, and it is not allowed by the config.';

      case UpdateResultValue::CannotApplyPatch:
        return 'The update failed because a patch included in your project could not be applied.';
    }
    return 'Unknown reason';
  }

  /**
   * {@inheritdoc}
   */
  public function getUpdateSuggestionsFromPackageAndOutput(string $package_name, string $console_output): array {
    $suggestions = [];
    $status = $this->getUpdateStatusFromOutput($console_output);
    switch ($status) {
      case UpdateResultValue::NewPluginNotAllowed:
        $suggestions[] = 'Do the update manually';
        $suggestions[] = 'Push an update where the plugin is allowed either together with a manual update, or as a separate PR.';
        break;

      case UpdateResultValue::CannotApplyPatch:
        $suggestions[] = 'Check if the patch is still needed, or if it has been committed upstream.';
        $suggestions[] = 'If the patch is still needed, check if the patch still applies.';
        $suggestions[] = 'If the patch does not apply, fix the patch, and ideally collaborate with the upstream project to get it merged.';
        break;
    }
    return $suggestions;
  }

}
